"""APC Power Distribution Unit control object."""

import logging
from datetime import datetime

from pysnmp.entity.rfc3413.oneliner import cmdgen
from pysnmp.proto import rfc1902
from retrying import retry, RetryError


class APC(object):
    """APC Power Distribution Unit."""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods
    # This class requires more attributes and public methods to cover the
    # functionality of the device.

    ERROR_MSG = "SNMP {} of {} on {} failed."

    SNMP_VERSION_2_2C = 1
    SNMP_PORT = 161
    SNMP_TIMEOUT = 1.5  # 1.5 seconds
    SNMP_RETRIES = 2

    MAX_STOP_DELAY = 15000  # 15 seconds
    INTER_RETRY_WAIT = 500  # 0.5 seconds

    # APC PowerNet-MIB Base OID
    BASE_OID = (1, 3, 6, 1, 4, 1, 318, 1, 1, 26)

    # Static readonly data
    # PowerNet-MIB::rPDU2IdentName.1
    Q_NAME = BASE_OID + (2, 1, 3, 1)
    # PowerNet-MIB::rPDU2IdentLocation.1
    Q_LOCATION = BASE_OID + (2, 1, 4, 1)
    # PowerNet-MIB::rPDU2IdentHardwareRev.1
    Q_HARDWARE_REV = BASE_OID + (2, 1, 5, 1)
    # PowerNet-MIB::rPDU2IdentFirmwareRev.1
    Q_FIRMWARE_REV = BASE_OID + (2, 1, 6, 1)
    # PowerNet-MIB::rPDU2IdentDateOfManufacture.1
    Q_MANUFACTURE_DATE = BASE_OID + (2, 1, 7, 1)
    # PowerNet-MIB::rPDU2IdentModelNumber.1
    Q_MODEL_NUMBER = BASE_OID + (2, 1, 8, 1)
    # PowerNet-MIB::rPDU2IdentSerialNumber.1
    Q_SERIAL_NUMBER = BASE_OID + (2, 1, 9, 1)
    # PowerNet-MIB::rPDU2DevicePropertiesNumOutlets.1
    Q_NUM_OUTLETS = BASE_OID + (4, 2, 1, 4, 1)
    # PowerNet-MIB::rPDU2DevicePropertiesNumSwitchedOutlets.1
    Q_NUM_SWITCHED_OUTLETS = BASE_OID + (4, 2, 1, 5, 1)
    # PowerNet-MIB::rPDU2DevicePropertiesNumMeteredOutlets.1
    Q_NUM_METERED_OUTLETS = BASE_OID + (4, 2, 1, 6, 1)
    # PowerNet-MIB::rPDU2DevicePropertiesMaxCurrentRating.1
    Q_MAX_CURRENT_RATING = BASE_OID + (4, 2, 1, 9, 1)
    # PowerNet-MIB::rPDU2PhaseStatusVoltage.1
    Q_PHASE_VOLTAGE = BASE_OID + (6, 3, 1, 6, 1)

    # Dynamic readonly data
    # PowerNet-MIB::rPDU2PhaseStatusLoadState.1
    Q_PHASE_LOAD_STATE = BASE_OID + (6, 3, 1, 4, 1)
    # PowerNet-MIB::rPDU2PhaseStatusCurrent.1
    Q_PHASE_CURRENT = BASE_OID + (6, 3, 1, 5, 1)
    # PowerNet-MIB::rPDU2DeviceStatusPower.1
    Q_POWER = BASE_OID + (4, 3, 1, 5, 1)
    # PowerNet-MIB::rPDU2SensorTempHumidityStatusType.1
    Q_SENSOR_TYPE = BASE_OID + (10, 2, 2, 1, 5, 1)
    # PowerNet-MIB::rPDU2SensorTempHumidityStatusName.1
    Q_SENSOR_NAME = BASE_OID + (10, 2, 2, 1, 3, 1)
    # PowerNet-MIB::rPDU2SensorTempHumidityStatusCommStatus.1
    Q_SENSOR_COMM_STATUS = BASE_OID + (10, 2, 2, 1, 6, 1)
    # PowerNet-MIB::rPDU2SensorTempHumidityStatusTempF.1
    Q_SENSOR_TEMP_F = BASE_OID + (10, 2, 2, 1, 7, 1)
    # PowerNet-MIB::rPDU2SensorTempHumidityStatusTempC.1
    Q_SENSOR_TEMP_C = BASE_OID + (10, 2, 2, 1, 8, 1)
    # PowerNet-MIB::rPDU2SensorTempHumidityStatusTempStatus.1
    Q_SENSOR_TEMP_STATUS = BASE_OID + (10, 2, 2, 1, 9, 1)
    # PowerNet-MIB::rPDU2SensorTempHumidityStatusRelativeHumidity.1
    Q_SENSOR_HUMIDITY = BASE_OID + (10, 2, 2, 1, 10, 1)
    # PowerNet-MIB::rPDU2SensorTempHumidityStatusHumidityStatus.1
    Q_SENSOR_HUMIDITY_STATUS = BASE_OID + (10, 2, 2, 1, 11, 1)
    # PowerNet-MIB::rPDU2OutletSwitchedStatusName.24
    Q_OUTLET_NAME = BASE_OID + (9, 2, 3, 1, 3)  # Requires outlet number
    # PowerNet-MIB::rPDU2OutletSwitchedStatusState.24
    Q_OUTLET_STATUS = BASE_OID + (9, 2, 3, 1, 5)  # Requires outlet number

    # Dynamic readwrite data
    # PowerNet-MIB::rPDU2SensorTempHumidityConfigName.1
    Q_SENSOR_NAME_RW = BASE_OID + (10, 2, 1, 1, 3, 1)
    # PowerNet-MIB::rPDU2OutletSwitchedConfigName.24
    Q_OUTLET_NAME_RW = BASE_OID + (9, 2, 1, 1, 3)  # Requires outlet number
    # PowerNet-MIB::rPDU2OutletSwitchedControlCommand.24
    Q_OUTLET_COMMAND_RW = BASE_OID + (9, 2, 4, 1, 5)  # Requires outlet number

    # Lookups
    LOAD_STATES = ["", "lowLoad", "normal", "nearOverload", "overload"]
    SENSOR_TYPES = [
        "",
        "temperatureOnly",
        "temperatureHumidity",
        "commsLost",
        "notInstalled",
    ]
    COMM_STATUS_TYPES = ["", "notInstalled", "commsOK", "commsLost"]
    SENSOR_STATUS_TYPES = [
        "",
        "notPresent",
        "belowMin",
        "belowLow",
        "normal",
        "aboveHigh",
        "aboveMax",
    ]
    OUTLET_STATUS_TYPES = ["", "off", "on"]

    def __init__(
        self, hostname_or_ip_address, public_community, private_community
    ):
        """Create an APC object.

        :param hostname_or_ip_address: hostname or ip address of PDU
        :param public_community: public community string
        :param private_community: private community string
        :return: APC object
        """
        self._host = hostname_or_ip_address
        self._vendor = "APC"

        self._transport = cmdgen.UdpTransportTarget(
            (self._host, self.SNMP_PORT),
            timeout=self.SNMP_TIMEOUT,
            retries=self.SNMP_RETRIES,
        )
        self._public = cmdgen.CommunityData(
            public_community, mpModel=self.SNMP_VERSION_2_2C
        )
        self._private = cmdgen.CommunityData(
            private_community, mpModel=self.SNMP_VERSION_2_2C
        )

        # Generic information (static)
        self._identification = self.__get(self.Q_NAME)
        self._location = self.__get(self.Q_LOCATION)
        self._hardware_rev = self.__get(self.Q_HARDWARE_REV)
        self._firmware_rev = self.__get(self.Q_FIRMWARE_REV)
        self._manufacture_date = datetime.strptime(
            str(self.__get(self.Q_MANUFACTURE_DATE)), "%m/%d/%Y"
        )
        self._model_number = self.__get(self.Q_MODEL_NUMBER)
        self._serial_number = self.__get(self.Q_SERIAL_NUMBER)

        # Device status (static)
        self._num_outlets = self.__get(self.Q_NUM_OUTLETS)
        self._num_switched_outlets = self.__get(self.Q_NUM_SWITCHED_OUTLETS)
        self._num_metered_outlets = self.__get(self.Q_NUM_METERED_OUTLETS)
        self._max_current = self.__get(self.Q_MAX_CURRENT_RATING)

        # Phase status (static)
        self._power_factor = 100
        self._current_factor = 10
        self._phase_voltage = int(self.__get(self.Q_PHASE_VOLTAGE))

        self._use_centigrade = False

        self._logger = logging.getLogger(__name__)
        self._logger.addHandler(logging.NullHandler())

    @property
    def host(self):
        """Hostname or IP Address of PDU.

        :return: PDU hostname or ip address
        """
        self._logger.info("Host: %s", self._host)
        return self._host

    @property
    def vendor(self):
        """Vendor/Manufacturer of PDU.

        :return: PDU vendor/manufacturer
        """
        self._logger.info("Vendor: %s", self._vendor)
        return self._vendor

    @property
    def identification(self):
        """Identification string.

        :return: PDU identification
        """
        self._logger.info("Identification: %s", self._identification)
        return self._identification

    @property
    def location(self):
        """Location of the PDU.

        :return: PDU location
        """
        self._logger.info("Location: %s", self._location)
        return self._location

    @property
    def hardware_revision(self):
        """Hardware revision.

        :return: PDU hardware revision
        """
        revision = str(self._hardware_rev)
        self._logger.info("Hardware revision: %s", revision)
        return revision

    @property
    def firmware_revision(self):
        """Firmware revision.

        :return: PDU firmware revision
        """
        revision = str(self._firmware_rev)
        self._logger.info("Firmware revision: %s", revision)
        return revision

    @property
    def date_of_manufacture(self):
        """Date of manufacture.

        :return: PDU date of manufacture
        """
        self._logger.info(
            "Date of Manufacture: %s", str(self._manufacture_date)
        )
        return self._manufacture_date

    @property
    def model_number(self):
        """Model number.

        :return: PDU model number
        """
        model = str(self._model_number)
        self._logger.info("Model number: %s", model)
        return model

    @property
    def serial_number(self):
        """Return the serial number.

        :return: PDU serial number
        """
        serial = str(self._serial_number)
        self._logger.info("Serial number: %s", serial)
        return serial

    @property
    def num_outlets(self):
        """Return the number of outlets in the PDU.

        :return: total number of outlets in the PDU
        """
        num = int(self._num_outlets)
        self._logger.info("Number of outlets: %d", num)
        return num

    @property
    def num_switched_outlets(self):
        """Return the number of switched outlets in the PDU.

        :return: number of switched outlets in the PDU
        """
        num = int(self._num_switched_outlets)
        self._logger.info("Number of switched outlets: %d", num)
        return num

    @property
    def num_metered_outlets(self):
        """Return the number of metered outlets in the PDU.

        :return: number of metered outlets in the PDU
        """
        num = int(self._num_metered_outlets)
        self._logger.info("Number of metered outlets: %d", num)
        return num

    @property
    def max_current(self):
        """Maximum current for the PDU.

        :return: maximum current for the PDU
        """
        current = int(self._max_current)
        self._logger.info("Maximum current: %d", current)
        return current

    @property
    def voltage(self):
        """Line voltage of the PDU.

        :return: PDU line voltage
        """
        voltage = int(self._phase_voltage)
        self._logger.info("Line voltage: %d", voltage)
        return voltage

    @property
    def load_state(self):
        """Load state of the PDU.

        :return: one of ['lowLoad', 'normal', 'nearOverload', overload']
        """
        state = int(self.__get(self.Q_PHASE_LOAD_STATE))
        self._logger.info("Load state: %s", self.LOAD_STATES[state])
        return self.LOAD_STATES[state]

    @property
    def current(self):
        """Return the current utilization of the PDU.

        :return: current, in amps
        """
        current = float(self.__get(self.Q_PHASE_CURRENT) / self._current_factor)
        self._logger.info("Current: %.2f", current)
        return current

    @property
    def power(self):
        """Return the power utilization of the PDU.

        :return: power, in kW
        """
        power = float(self.__get(self.Q_POWER) / self._power_factor)
        self._logger.info("Power: %.2f", power)
        return power

    @property
    def is_sensor_present(self):
        """Determine if a sensor is present on the PDU.

        :return: Is the sensor present?
        """
        state = self.__get(self.Q_SENSOR_TYPE)
        present = 1 < int(state) < 3
        self._logger.info("Sensor present: %s", str(present))
        return present

    @property
    def sensor_name(self):
        """Name of the sensor.

        :return: name of the sensor
        """
        name = None
        if self.is_sensor_present:
            name = str(self.__get(self.Q_SENSOR_NAME))
        self._logger.info("Sensor name: %s", name)
        return name

    @sensor_name.setter
    def _set_sensor_name(self, name):
        """Name of the sensor.

        :param name: name of the sensor
        :return:
        """
        if self.is_sensor_present:
            self.__set(self.Q_SENSOR_NAME_RW, name)
            self._logger.info("Updating sensor name to: %s", name)

    @property
    def sensor_type(self):
        """Type of sensor.

        :return: type of sensor, one of
            ['temperatureOnly', 'temperatureHumidity', 'commsLost',
            'notInstalled']
        """
        index = 4
        if self.is_sensor_present:
            index = int(self.__get(self.Q_SENSOR_TYPE))
        self._logger.info("Sensor type: %s", self.SENSOR_TYPES[index])
        return self.SENSOR_TYPES[index]

    @property
    def sensor_comm_status(self):
        """Communication status of the sensor.

        :return: communication status of the sensor
        """
        index = 1
        if self.is_sensor_present:
            index = int(self.__get(self.Q_SENSOR_COMM_STATUS))
        self._logger.info(
            "Sensor communication status: %s", self.SENSOR_STATUS_TYPES[index]
        )
        return self.SENSOR_STATUS_TYPES[index]

    @property
    def use_centigrade(self):
        """Select between centigrade and fahrenheit.

        :return: using centigrade or not
        """
        self._logger.info("Use centigrade: %s", str(self._use_centigrade))
        return self._use_centigrade

    @use_centigrade.setter
    def use_centigrade(self, value):
        """Select between centigrade and fahrenheit.

        :param value: use centrigrade or not
        :return:
        """
        self._logger.info("Updating use centigrade to: %s", value)
        self._use_centigrade = value

    @property
    def temperature(self):
        """Temperature.

        :return: temperature
        """
        temp = 0.00
        if self.sensor_supports_temperature:
            if self._use_centigrade:
                temp = float(self.__get(self.Q_SENSOR_TEMP_C) / 10)
            else:
                temp = float(self.__get(self.Q_SENSOR_TEMP_F) / 10)
        self._logger.info("Temperature: %.2f", temp)
        return temp

    @property
    def humidity(self):
        """Relative humidity.

        :return: relative humidity
        """
        humid = 0.00
        if self.sensor_supports_humidity:
            humid = float(self.__get(self.Q_SENSOR_HUMIDITY))
        self._logger.info("Relative humidity: %.2f", humid)
        return humid

    @property
    def temperature_status(self):
        """Determine the status of the temperature sensor.

        :return: The status of the temperature sensor
        """
        index = 1
        if self.sensor_supports_temperature:
            index = self.__get(self.Q_SENSOR_TEMP_STATUS)
        self._logger.info(
            "Temperature sensor status: %s", self.SENSOR_STATUS_TYPES[index]
        )
        return self.SENSOR_STATUS_TYPES[index]

    @property
    def humidity_status(self):
        """Determine the status of the humidity sensor.

        :return: status of the humidity sensor
        """
        index = 1
        if self.sensor_supports_humidity:
            index = self.__get(self.Q_SENSOR_HUMIDITY_STATUS)
        self._logger.info(
            "Relative humidity sensor status: %s",
            self.SENSOR_STATUS_TYPES[index],
        )
        return self.SENSOR_STATUS_TYPES[index]

    def get_outlet_name(self, outlet):
        """Name of an outlet in the PDU.

        :param outlet: outlet number
        :return: name of the outlet
        """
        if 1 <= outlet <= self._num_outlets:
            name = str(self.__get(self.Q_OUTLET_NAME + (outlet,)))
            self._logger.info("Outlet number %d has name %s", outlet, name)
            return name
        raise IndexError(
            'Only {} outlets exist. "{}" is an invalid outlet.'.format(
                self._num_outlets, str(outlet)
            )
        )

    def set_outlet_name(self, outlet, name):
        """Update the name of an outlet in the PDU.

        :param outlet: outlet number
        :param name: outlet name
        :return:
        """
        if 1 <= outlet <= self._num_outlets:
            self.__set(self.Q_OUTLET_NAME_RW + (outlet,), name)
            self._logger.info(
                "Updating outlet number %d to new name %s", outlet, name
            )
        raise IndexError(
            'Only {} outlets exist. "{}" is an invalid outlet.'.format(
                self._num_outlets, str(outlet)
            )
        )

    def outlet_status(self, outlet):
        """Determine the status of the outlet in the PDU.

        :param outlet: outlet number
        :return: status of the outlet, one of ['on', 'off']
        """
        if 1 <= outlet <= self._num_outlets:
            state = self.__get(self.Q_OUTLET_STATUS + (outlet,))
            self._logger.info(
                "Outlet number %d has status %s",
                outlet,
                self.OUTLET_STATUS_TYPES[state],
            )
            return self.OUTLET_STATUS_TYPES[state]
        raise IndexError(
            'Only {} outlets exist. "{}" is an invalid outlet.'.format(
                self._num_outlets, str(outlet)
            )
        )

    def outlet_command(self, outlet, operation):
        """Send command to an outlet in the PDU.

        :param outlet: outlet number
        :param operation: one of ['on', 'off', 'reboot']
        :return: did the operation complete successfully?
        """
        valid_operations = ["on", "off", "reboot"]
        if operation not in valid_operations:
            raise ValueError(
                '"{}" is an invalid operation. Valid operations are: {}'.format(
                    str(operation), str(valid_operations)
                )
            )

        operations = {"on": 1, "off": 2, "reboot": 3}

        if 1 <= outlet <= self._num_outlets:
            self._logger.info(
                "Setting outlet %d to %s state", outlet, operation
            )
            self.__set(
                self.Q_OUTLET_COMMAND_RW + (outlet,), operations[operation]
            )

            try:
                if operation in ("on", "reboot"):
                    success = self.__wait_for_state(outlet, "on")
                else:
                    success = self.__wait_for_state(outlet, "off")
            except RetryError:
                # If the operation timed out, no determination of the result
                # can be made.
                success = False

            return success
        raise IndexError(
            'Only {} outlets exist. "{}" is an invalid outlet.'.format(
                self._num_outlets, str(outlet)
            )
        )

    @property
    def sensor_supports_temperature(self):
        """Determine if the sensor supports temperature measurements.

        :return: does the sensor support temperature measurements?
        """
        return self.is_sensor_present and "temp" in self.sensor_type.lower()

    @property
    def sensor_supports_humidity(self):
        """Determine if the sensor supports relative humidity measurements.

        :return: does the sensor support relative humidity measurements?
        """
        return self.is_sensor_present and "humid" in self.sensor_type.lower()

    # pylint: disable=no-self-argument
    # In order to use this method within the @retry decorator, this method
    # must be defined as such.
    def __retry_if_not_state(result):
        """Only keep retrying if the state is not what is expected.

        :return: negation of input
        """
        return not result

    @retry(
        stop_max_delay=MAX_STOP_DELAY,
        wait_fixed=INTER_RETRY_WAIT,
        retry_on_result=__retry_if_not_state,
    )
    def __wait_for_state(self, outlet, state):
        """Wait until state is hit.

        This will wait for MAX_STOP_DELAY with a inter-try delay of
        INTER_RETRY_WAIT.

        :param outlet: outlet number
        :param state: state to wait for
        :return: was the state hit?
        """
        return self.outlet_status(outlet) is state

    def __get(self, oid):
        """Get a specific value from an OID in the SNMP tree.

        :param oid: OID to get
        :returns: value from the specified OID
        """
        (error_indication, _, _, var_binds) = cmdgen.CommandGenerator().getCmd(
            self._public, self._transport, oid
        )
        if error_indication:
            raise RuntimeError(self.ERROR_MSG.format("get", oid, self._host))

        return var_binds[0][1]

    def __set(self, oid, value):
        """Set a specific value to an OID in the SNMP tree.

        :param oid: OID to set
        :param value: value to set
        """
        initial_value = self.__get(oid)
        new_value = self.__coerce_value(initial_value, value)

        (error_indication, _, _, var_binds) = cmdgen.CommandGenerator().setCmd(
            self._private, self._transport, (oid, new_value)
        )
        if error_indication:
            raise RuntimeError(self.ERROR_MSG.format("set", oid, self._host))

        return var_binds[0][1]

    @staticmethod
    def __coerce_value(initial_value, new_value):
        """Coerce the new_value to the same type as the initial_value.

        Unfortunately this is a bit of a workaround for the more elegant
        version:
        `return initial_value.__init__(str(new_value))`
        Utilizing that more elegant version yields an SmiError:
        MIB object ObjectIdentity((...)) is not OBJECT-TYPE (MIB not loaded?)

        :param initial_value: initial value from the device
        :param new_value: new value to set, coerced into the right type
        :return: new value, coerced into the right type
        """
        if isinstance(initial_value, rfc1902.Counter32):
            set_value = rfc1902.Counter32(str(new_value))
        elif isinstance(initial_value, rfc1902.Counter64):
            set_value = rfc1902.Counter64(str(new_value))
        elif isinstance(initial_value, rfc1902.Gauge32):
            set_value = rfc1902.Gauge32(str(new_value))
        elif isinstance(initial_value, rfc1902.Integer):
            set_value = rfc1902.Integer(str(new_value))
        elif isinstance(initial_value, rfc1902.Integer32):
            set_value = rfc1902.Integer32(str(new_value))
        elif isinstance(initial_value, rfc1902.IpAddress):
            set_value = rfc1902.IpAddress(str(new_value))
        elif isinstance(initial_value, rfc1902.OctetString):
            set_value = rfc1902.OctetString(str(new_value))
        elif isinstance(initial_value, rfc1902.TimeTicks):
            set_value = rfc1902.TimeTicks(str(new_value))
        elif isinstance(initial_value, rfc1902.Unsigned32):
            set_value = rfc1902.Unsigned32(str(new_value))
        else:
            raise RuntimeError("Unknown type: {}".format(type(initial_value)))

        return set_value



"""Cisco Switch control object."""

import logging
import warnings

from netaddr import EUI, mac_unix_expanded
from netmiko import ConnectHandler


def deprecated(func):
    """Decorate a deprecated function to warn when it is called.

    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    def new_func(*args, **kwargs):
        """Set the warning message."""
        warnings.warn(
            "Call to deprecated function {}.".format(func.__name__),
            category=DeprecationWarning,
        )
        return func(*args, **kwargs)

    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    new_func.__dict__.update(func.__dict__)

    return new_func


class CiscoSwitch(object):
    """Cisco switch control."""

    # pylint: disable=len-as-condition
    # Parsing command line output has uncertain lengths

    MAX_COMMAND_READ = 16

    CMD_VERSION = "sh version"
    CMD_VERSION_SIGNALS = ["BOOTLDR"]

    CMD_IPDT = "sh ip device track all"

    CMD_MAC_ADDRESS_TABLE = "sh mac address-table"

    CMD_POWER_OFF = "power inline never"
    CMD_POWER_ON = "power inline auto"
    CMD_POWER_LIMIT = "power inline auto max {0}"
    CMD_POWER_SHOW = "sh power inline {0}"

    CMD_CONFIGURE_INTERFACE = "int {0}"

    CMD_VLAN_MODE_ACCESS = "switchport mode access"
    CMD_VLAN_SET = "switchport access vlan {0}"
    CMD_VLAN_SHOW = "sh vlan"

    CMD_CARRIAGE_RETURN = "\n"

    PORT_NOTATION = {
        "fastethernet": "Fa",
        "gigabitethernet": "Gi",
        "tengigabitethernet": "Ten",
    }

    def __init__(self, hostname_or_ip_address, username, password):
        """Initialize the CiscoSwitch object.

        :param hostname_or_ip_address: host name or ip address
        :param username: username to login with
        :param password: password to use with username
        :return: CiscoSwitch object
        """
        self._host = hostname_or_ip_address
        self._version = None
        self._username = username
        self._password = password
        self._client = None

        self._logger = logging.getLogger(__name__)
        self._logger.addHandler(logging.NullHandler())

    def connect(self):
        """Connect to the switch.

        :return:
        """
        self._client = ConnectHandler(
            device_type="cisco_ios",
            ip=self._host,
            username=self._username,
            password=self._password,
        )

    def disconnect(self):
        """Disconnect from the switch.

        :return:
        """
        if self._client is not None:
            self._client.disconnect()
            self._client = None

    def enable(self, password):
        """Put the switch in enable mode.

        :param password: password to use to enable
        :return:
        """
        if not self.connected:
            return

        self._client.secret = password
        self._client.enable()

    @deprecated
    def set_terminal_length(self):
        """Set terminal length.

        :return:
        """
        pass

    def ipdt(self):
        """IP Device Tracking (IPDT) information.

        :return: IPDT information
        """
        if not self.connected:
            return None

        output = self._send_command(self.CMD_IPDT)
        return self._parse_ipdt_output(output)

    def mac_address_table(self, ignore_port=None):
        """MAC Address Table Information.

        :param ignore_port: port to ignore, e.g. Gi1/0/48
        :return: ARP information
        """
        if not self.connected:
            return None

        output = self._send_command(self.CMD_MAC_ADDRESS_TABLE)
        return self._parse_mac_address_table_output(
            output, ignore_port=ignore_port
        )

    def poe_on(self, port):
        """Enable a port for POE.

        :param port: port to enable POE on, e.g. Gi1/0/1
        :return: True if the command succeeded, False otherwise
        """
        if not self.connected:
            return False

        port = self._shorthand_port_notation(port)
        cmds = [self.CMD_CONFIGURE_INTERFACE.format(port), self.CMD_POWER_ON]
        self._send_config(cmds)

        verify = self._send_command(self.CMD_POWER_SHOW.format(port))
        matches, _, _ = CiscoSwitch._verify_poe_status(verify, port, "on")
        return matches

    def poe_off(self, port):
        """Disable a port for POE.

        :param port: port to disable POE on, e.g. Gi1/0/1
        :return: True if the command succeeded, False otherwise
        """
        if not self.connected:
            return False

        port = self._shorthand_port_notation(port)
        cmds = [self.CMD_CONFIGURE_INTERFACE.format(port), self.CMD_POWER_OFF]
        self._send_config(cmds)

        verify = self._send_command(self.CMD_POWER_SHOW.format(port))
        matches, _, _ = CiscoSwitch._verify_poe_status(verify, port, "off")
        return matches

    def poe_limit(self, port, milliwatts_limit):
        """Enable a port for POE, but limit the maximum wattage.

        :param port: port to enable POE on, e.g. Gi1/0/1
        :param milliwatts_limit: the maximum wattage,
            given in milliwatts, e.g. 15400. Cisco documentation
            gives 4000 to 30000 as the valid range.
        """
        if not self.connected:
            return False

        port = self._shorthand_port_notation(port)
        cmds = [
            self.CMD_CONFIGURE_INTERFACE.format(port),
            self.CMD_POWER_LIMIT.format(milliwatts_limit),
        ]
        self._send_config(cmds)

        verify = self._send_command(self.CMD_POWER_SHOW.format(port))
        matches, _, _ = CiscoSwitch._verify_poe_status(verify, port, "on")
        return matches

    def is_poe(self, port):
        """Get the POE state for a port.

        :param port: port to determine state of, e.g. Gi1/0/1
        :return: milliwatts_limit (a non-zero integer), evaluating to True if
            POE is enabled, else 0 (False).
        """
        if not self.connected:
            return 0

        port = self._shorthand_port_notation(port)
        verify = self._send_command(self.CMD_POWER_SHOW.format(port))
        _, poe, limit = CiscoSwitch._verify_poe_status(verify, port, "unknown")
        return 0 if poe == "off" else limit

    def change_vlan(self, port, vlan):
        """Change the VLAN assignment on a port.

        :param port: port to change VLAN assignment on, e.g. Gi1/0/1
        :param vlan: VLAN id
        :return: True if the command succeeded, False otherwise
        """
        if not self.connected:
            return False

        port = self._shorthand_port_notation(port)
        cmds = [
            self.CMD_CONFIGURE_INTERFACE.format(port),
            self.CMD_VLAN_MODE_ACCESS,
            self.CMD_VLAN_SET.format(int(vlan)),
        ]
        self._send_config(cmds)

        verify = self._send_command(self.CMD_VLAN_SHOW)
        matches, _ = CiscoSwitch._verify_vlan_status(verify, port, int(vlan))
        return matches

    def vlan(self, port):
        """Get the VLAN assignment on a port.

        :param port: port to determine VLAN assignment on, e.g. Gi1/0/1
        :return: VLAN id
        """
        if not self.connected:
            return -1

        port = self._shorthand_port_notation(port)
        verify = self._send_command(self.CMD_VLAN_SHOW)
        _, vlan = CiscoSwitch._verify_vlan_status(verify, port, 0)
        return vlan

    @property
    def version(self):
        """Retrieve the Cisco IOS version.

        :return: The Cisco IOS version.
        """
        if not self.connected:
            return None

        if self._version is None:
            output = self._send_command(self.CMD_VERSION)
            self._version = self._parse_version_output(output)

        return self._version

    @property
    def host(self):
        """Retrieve the IP address or hostname of the switch.

        :return: IP address or hostname of the switch
        """
        return self._host

    @property
    def connected(self):
        """Switch connection status.

        :return: has a connection to the switch been made?
        """
        if self._client is None:
            return False

        self._send_command(self.CMD_CARRIAGE_RETURN)

        active_ssh = True
        try:
            self._client.find_prompt()
        except ValueError:
            active_ssh = False

        return active_ssh

    def _shorthand_port_notation(self, port):
        """Shorthand port notation.

        Takes a port name, such as Gi1/0/48 or GigabitEthernet1/0/48 and
        returns the shorthand notation.

        :param port: port name
        :return: shorthand port name
        """
        if port is None:
            return ""

        lower = port.lower()
        output = port

        if any(lower.find(port) == 0 for port in self.PORT_NOTATION):
            for item in self.PORT_NOTATION.items():
                if lower.startswith(item[0]):
                    output = lower.replace(item[0], item[1])
                    break

        return output

    def _send_command(self, command):
        """Send command.

        Sends a command to the switch.

        :param command: command to send
        :return: output of the command
        """
        read_buffer = ""
        if self._client is None:
            return read_buffer

        self._client.clear_buffer()
        read_buffer = self._client.send_command(command)

        return read_buffer

    def _send_config(self, configs):
        """Configure switch.

        Sends a series of configuration options to the switch.

        :param configs: config commands
        """
        if self._client is None:
            return

        self._client.clear_buffer()
        self._client.send_config_set(configs)

    @staticmethod
    def _parse_version_output(output):
        """Version parsing.

        :param output: the output of the command
        :return: version string
        """
        lines = [line.strip() for line in output.splitlines()]
        version = ""
        search = True
        while len(lines) > 0 and search:
            line = lines.pop()

            if line.find("Cisco IOS") < 0:
                continue

            version_info = line.split(",")

            if len(version_info) > 1:
                version = version_info[2].strip()
                version = version[8:]
                search = False

        return version

    def _parse_mac_address_table_output(self, output, ignore_port=None):
        """MAC Address Table parsing.

        Parses the output of 'sh mac address-table' command.

        :param output: the output of the command
                      Mac Address Table
            -------------------------------------------
            Vlan    Mac Address       Type        Ports
            ----    -----------       --------    -----
             All    0100.0ccc.cccc    STATIC      CPU
             <snip>
             All    ffff.ffff.ffff    STATIC      CPU
             601    000b.7866.5240    DYNAMIC     Gi1/0/48
             601    0013.20fe.56b4    DYNAMIC     Gi1/0/35
             <snip>
             Total Mac Addresses for this criterion: 59
        :param ignore_port: ignore this port
        :return: list of dicts containing device connection metrics
        """
        lookup = []
        lines = [line.strip() for line in output.splitlines()]
        lines.append("")
        while len(lines) > 0:
            line = lines.pop(0)

            # Table entries will always have a '.' for the MAC address.
            # If there isn't one it's not a row we care about.
            if line.find(".") < 0:
                continue

            values = [entry for entry in line.split() if entry]

            # Ignore non-physical ports
            port_types = self.PORT_NOTATION.values()
            if all(
                values[3].lower().find(port.lower()) != 0 for port in port_types
            ):
                continue

            # If the ignore_port is specified and is the port in question,
            # ignore it.
            ignore_port = self._shorthand_port_notation(ignore_port)
            if ignore_port in ("", values[3]):
                continue

            lookup.append(
                {
                    "mac": EUI(values[1], dialect=mac_unix_expanded),
                    "interface": self._shorthand_port_notation(values[3]),
                }
            )
        return sorted(lookup, key=lambda k: k["interface"])

    def _parse_ipdt_output(self, output):
        """IPDT output parsing.

        Parses the output of the `show ip device track` command.

        :param output: the output of the command
            IP Device Tracking = Enabled
            IP Device Tracking Probe Count = 3
            IP Device Tracking Probe Interval = 30
            IP Device Tracking Probe Delay Interval = 0
            -------------------------------------------------------------------
              IP Address     MAC Address   Vlan  Interface              STATE
            -------------------------------------------------------------------
            192.168.1.12     6cec.eb68.c86f  601  GigabitEthernet1/0/14  ACTIVE
            192.168.1.15     6cec.eb67.836c  601  GigabitEthernet1/0/12  ACTIVE
            <snip>

            Total number interfaces enabled: 47
            Enabled interfaces:
             Gi1/0/1, Gi1/0/2, Gi1/0/3, Gi1/0/4, Gi1/0/5, Gi1/0/6, Gi1/0/7,
             <snip>
        :return: list of dicts containing device connection metrics
        """
        lookup = []
        lines = [line.strip() for line in output.splitlines()]
        lines.append("")
        while len(lines) > 0:
            line = lines.pop(0)

            # Table entries will always have a '.' for the MAC address.
            # If there isn't one it's not a row we care about.
            if line.find(".") < 0:
                continue

            values = [entry for entry in line.split() if entry]

            # Ignore any 'INACTIVE' entries, meaning that there hasn't been
            # traffic from that MAC Address is has likely been unplugged.
            if values[4] == "INACTIVE":
                continue

            lookup.append(
                {
                    "ip": values[0],
                    "mac": EUI(values[1], dialect=mac_unix_expanded),
                    "interface": self._shorthand_port_notation(values[3]),
                }
            )
        return sorted(lookup, key=lambda k: k["interface"])

    @staticmethod
    def _verify_poe_status(output, port, state):
        """Verify that the given port is in the given state.

        :param output: the output of the command
            Interface Admin  Oper       Power   Device              Class Max
                                        (Watts)
            --------- ------ ---------- ------- ------------------- ----- ----
            Gi2/0/16  auto   on         15.4    Ieee PD             0     30.0

            Interface  AdminPowerMax   AdminConsumption
                         (Watts)           (Watts)
            ---------- --------------- --------------------

            Gi2/0/16              30.0                 15.4
        :param port: port to check
        :param state: expected state
        :return: (True, actual_state, milliwatts_limit) if the port is in the
            expected state, (False, actual_state, milliwatts_limit) otherwise.
        """
        if "on" in state:
            state = "auto"

        matches = False
        lines = [line.strip() for line in output.splitlines()]
        lines.append("")
        actual_state = "unknown"
        milliwatts_limit = 0
        while len(lines) > 0:
            line = lines.pop(0)

            # Table entries will always have a '/' for the interface.
            # If there isn't one it's not a row we care about.
            if line.find("/") < 0:
                continue

            values = [entry for entry in line.split() if entry]

            # Is this the port of interest?
            if values[0] != port:
                continue

            matches = state in values[1]
            actual_state = values[1]
            # The last column, "Max", gives Watts to the to tenths spot
            milliwatts_limit = int(float(values[-1]) * 1000.0)
            break

        return matches, actual_state, milliwatts_limit

    @staticmethod
    def _verify_vlan_status(output, port, vlan):
        """Verify that the given port is assigned to the given vlan.

        :param output:
            VLAN Name                             Status    Ports
            ---- -------------------------------- --------- ------------------
            1    default                          active    Te1/0/1, Te1/0/2
            701  NET-701                          active    Gi1/0/1, Gi1/0/2
            702  NET-702                          active    Gi1/0/8
            703  NET-703                          active    Gi1/0/7, Gi1/0/10
            704  NET-704                          active    Gi1/0/5, Gi1/0/9
            705  NET-705                          active

        :param port: port to check
        :param vlan: expected vlan
        :return: (True, actual vlan #) if the port is assigned to the expected
        vlan, (False, actual vlan #) otherwise
        """
        matches = False
        lines = [line.strip() for line in output.splitlines()]
        lines.append("")
        actual_vlan = -1
        current_vlan = -1
        while len(lines) > 0:
            line = lines.pop(0).replace(",", "")

            # Table entries will have a '/' for the ports.
            # If there isn't one it's not a row we care about, even if it's a
            # vlan with no ports assigned to it.
            if line.find("/") < 0:
                continue

            values = [entry for entry in line.split() if entry]
            if "active" in values:
                current_vlan = int(values[0])
                ports = values[3:]
            else:
                ports = values

            # Is the port of interest in the list of ports?
            if port not in ports:
                continue

            actual_vlan = current_vlan
            if actual_vlan == vlan:
                matches = True

        return matches, actual_vlan
